#!/bin/sh
# the next line restarts using tclsh\
exec tclsh "$0" "$@"

# Observation circumstance computer prototype.
# Current version (see "ident") computes real hsun using the local machine time
# (precise time should be set! Use XNTPD for example)
# and the longitude and latitude supplied by SV circumstance monitor scenario 
# It still supports the 2-minutes day simulation if set in test-mode (see below)
#
# Astronomical computations here are made using formulii given in two books:
# 1. See Astronomija s personal'nym compjuterom, Montenbruk, Pfleger, Moskva, 
# "Mir", 1993, abbreviated here as APC,
# 2. Astronomicheskie formuly dlja kal'kuljatora, Mees, Moskva, "Mir", 1988,
# abbreviated here as AFK.
#
# Apart from standard general commands (described in commands.tex in section
# General commands) the following commands are supported:
#
# SET LONGITUDE="+-DDD.DDDD" LATITUDE="+-DD.DDDD" (in degrees, positive for E,N)
# 		to override wrong defaults.
# SET MODE=TEST|REAL - to switch between simulated 2-minutes winter day and real
# computations
# GET UT - returns current UT as "YYYY MM DD hh mm ss" with 1sec precision
# GET JD - returns current Julian Day
# GET SUNRA SUNDEC - returns current equatorial coordinates of Sun in degrees
# GET HSUN - returns current altitude of Sun above horizon in degrees
#

#
# Port number:
set port 1234
# Identification:
set ident "circ.tcl v0.02"


# Default geographic coordinates:
set longitude  [expr 37 + 42./60.]
set latitude [expr 55. + 45./60]

# Mode: test/real
set realmode true

# Refresh period in seconds:
set refresh_period 5

# -------------- COMPUTABLE global variables
# Data: empty here
set data "-"

# UT string "YYYY MM DD hh mm ss"
set ut "0000 00 00 00 00 00"

# Julian day:
set jd 0

# Sun coordinates in degrees:
set sunra 0; set sundec 0

# Sun horizontal coordinates:
set azsun 0; set hsun 0

#
# @return UT as a string YYYY MM DD hh mm ss
# ATTENTION! Machine local time and - ESPECIALLY - timezone should be correct!
#
proc get_ut {} {
	global ut
	return [ set ut [\
	clock format [clock seconds] -format "%Y %m %d %H %M %S" -gmt true ]]
}

#
# Compute Julian day: see AFK, p.27
# @param ut UT date and time in format returned by get_ut{}
# @return Julian day (conventional, i.e. NOT with 2400000 subtracted)
#
proc get_jd ut {
	global jd
	# convert ut into list:
	set ut [split $ut]
	set y "[lindex $ut 0].0"; set m "[lindex $ut 1].0"
	set d "[lindex $ut 2].0"
	set dd [expr $d + [lindex $ut 3].0/24. + [lindex $ut 4].0/24./60 +\
		[lindex $ut 5].0/24./3600]
	puts "dd=$dd"
	if {$m<=2} {set m [expr $m+12]; set y [expr $y-1]}
	if {[expr $y+$m/100+$d/10000]>=1582.1015} {
		set a [expr int($y/100.)]; 
		set b [expr 2. - $a + int($a/4.)]
	} else { set b 0}
	set jd [expr int(365.25*$y - 0.75*($y<0)) + int(30.6001 * ($m+1)) + $dd + \
		1720994.5 + $b]
	puts "JD($ut)=$jd"
	return $jd
}

#
# Compute greenwich stellar time for 0hUT of given JD (AFK p.37)
# @param jd Julian date
# @return st in hours
#
proc get_st0 jd {
	# JD for 0hUT:
	set jd0 [expr int($jd+0.5)-0.5]
	set t [expr ($jd0 - 2415020.0)/36525]
	# Greenwich time for 0hUT in fractional hours:
	# set theta0 [expr 6.6460656 + 2400.051262 * $t + 0.00002581 * $t*$t]
	# Greenwich time for 0hUT in fractional days:
	set theta0 [expr 0.276919398 + 100.0021359 * $t + 0.000001075 *$t*$t]
	return [expr 24. * [frac $theta0]]
}

# For mini_sun:
# returns fractional part of argument
proc frac x {
	set x [expr $x - int($x)] 
	if {$x<0} {return [expr $x+1]} {return $x}
}
#
# Compute approximate equatorial coordinates of Sun from JD with prec.~1arcmin
# @return [list sunra sundec] in hours and degrees, respectively
# See APC, p.48
#
proc mini_sun jd {
	global sunra sundec
	# Time in julian centuries:
	set t [expr ($jd-2451545.0)/36525.0]
	puts "t=$t"
	set p2 6.283185307;	set coseps 0.91748; set sineps 0.39778
	set m [expr $p2 * [frac [expr 0.993133 + 99.997361 * $t]]]
	set dl [expr 6893.0 * sin($m) + 72.0*sin(2*$m)]
	set l [expr $p2 * [frac [expr 0.7859453 + $m/$p2 + (6191.2*$t+$dl)/1296.e3]]]
	set sl [expr sin($l)]
	set x [expr cos($l)]; set y [expr $coseps*$sl]; set z [expr $sineps*$sl]
	set rho [expr sqrt(1.0-$z*$z)]
	set sundec [expr (360.0/$p2)*atan($z/$rho)]
	set sunra [expr (48./$p2)*atan($y/($x+$rho))]
	if {$sunra<0} {set sunra [expr $sunra+24.0]}
#	puts "Sun (jd=$jd) RA=$sunra deg DEC=$sundec deg"
	return [list $sunra $sundec]
}

#
# Convert equatorial coordinates HA,DEC in horizontal A,h
# @param ha hour angle in hours
# @param dec declination in degrees
# @return [list azimuth altitude] in degrees
#
proc hadec2azalt {ha dec} {
	global latitude
	set pi 3.14159265358979323
	set ha [expr $ha*$pi/12.]
	set dec [expr $dec*$pi/180.]
	set fi [expr $latitude*$pi/180.]
	set sina [expr sin($ha)]
	set cosa [expr cos($ha)*sin($fi) - tan($dec)*cos($fi)]
	set sinh [expr sin($fi)*sin($dec) + cos($fi)*cos($dec)*cos($ha)]
	set a [expr atan2($sina,$cosa)*180/$pi]
	set h [expr asin($sinh)*180/$pi]
	return [list $a $h]
}

#
# Convert degrees in hh mm ss:
#
proc deg2hms deg {
	if {$deg<0 || $deg>=360} {set deg [expr 360.0 * [frac [expr $deg/360.0] ]]}
	return [deg2dms [expr $deg/15.]]
}
#
# Convert degrees in dd mm ss.ss:
#
proc deg2dms deg {
	set sig [expr 1-2*($deg<0)]
	set deg [expr abs($deg)]
	set d [expr int($deg)]
	set m [expr int([frac $deg]*60.0)]
	set s [expr ($deg-$d-$m/60.)*3600.0]
	return [list [expr $d*$sig] $m $s]
}
#
# Convert dd mm ss into fractional degrees or hours
#
proc dms2deg {dd mm ss} {
	set r [expr abs($dd) + abs($mm)/60. + abs($ss)/3600.]
	set sig [expr 1. - 2*[string match -* $dd]]
	return [expr $sig*$r]
}
#
# Calculate the altitude of Sun above horizon (no refraction accounted)
#
proc get_azaltsun {} {
	global realmode hsun azsun
	if {$realmode} {
puts "RA(Saturn)=[set ra [dms2deg 10 57 35.681]] hours"
puts "Dec(Saturn)=[set dec [expr 8 + 25/60. + 58.10/3600.]] deg"
set st [expr [get_st0 [get_jd "1978 11 13 4 34 00"]] + [dms2deg 4 34 00]*1.002737908]
puts "St0=$st hours"
set h [expr $st - $longitude - $ra]
puts "HA(Saturn)=$h"
puts "Az h = [hadec2azalt $h $dec]"
		
	} else {
		# Test-mode: we model Sun altitude as a sine-wave
		set per 240; set amp 45; set zerol -20; set pi 3.14159265358979323
		return [expr $zerol+$amp*sin(fmod([clock seconds], $per)*2*$pi/$per)]
	}
}

# Last data update time in seconds-1969:
set refresh_time 0

# Recompute all parameters:
#
proc compute {} {
	global refresh_time refresh_period
	# If data are still fresh - return:
	if {[clock seconds]-$refresh_time<$refresh_period} return
	get_ut
	get_jd $::ut
	set refresh_time [clock seconds]
}

proc cmd_get pars {
	# global array of parameter values cmp_pvar(param)=value and comp.status:
	global cmp_pvar cmp_status
	if {$pars!="ident"} compute
	foreach p $pars {
		switch $p {
			"ident" - "data" - "hsun" - "ut"
				{ upvar #0 $p par; set cmp_pvar($p) $par }
			default
				{ set cmp_status ERSYN; break }
		}
	}
}

# setting the parameters from list p1 v1 p2 v2... sent by name for upvar ref.
proc cmd_set pv_listname {
	global cmp_status longitude latitude realmode
	upvar $pv_listname pv_list
	array set pvar $pv_list
	# - would be enough if pvar is made global, trusting input completely. 
	foreach pname [array names pvar] {
		switch $pname {
			"longitude"
				{
					# Check range:
					if {$pvar($pname)>=-180 && $pvar($pname)<=+180} { 
						set longitude $pvar($pname) 
					} else {
						set cmp_status ERANG
						break
					}
				}
			"latitude"
				{
					# Check range:
					if {$pvar($pname)>=-90 && $pvar($pname)<=+90} { 
						set latitude $pvar($pname) 
					} else {
						set cmp_status ERANG
						break
					}
				}
			"mode"
				{
					if {[string tolower $pvar($pname)]=="real"} {
						set realmode 1
					} elseif {[string tolower $pvar($pname)]=="test"} {
						set realmode 0
					} else {
						set cmp_status ERANG
						break
					}
				}
			default
				{
					set cmp_status ERSYN
					break
				}
		}
	}
}

proc cmd_init {} {
	global cmp_status
	# read CFG file?
	set cmp_status READY
}

proc cmd_go_local {} {
	global cmp_status
	# enable accepting local console commands
	set cmp_status LOCAL
}

proc cmd_park {} {
	global cmp_status cmp_pvar
	# do nothing?
	set cmp_status PARKED
}

proc cmd_go_online {} {
	global cmp_status
	# disable local console commands
	set cmp_status READY
}

source compcmd.tcl

#open_server $port
#while 1 { after 1000; puts "hsun=[format %-3.1f [compute_hsun]] deg" }
#puts [get_jd "1957 10 4 19 26 24"]; puts [get_jd "333 1 27 12 00 00"]
#puts [get_jd "-584 05 28 15 07 12"]; 
#puts [get_jd "1900 01 01 00 00 00"];
#puts [get_jd [get_ut]]
#puts [mini_sun [get_jd "1978 11 12 00 00 00"]]
#puts "RA in H M S: [deg2dms $sunra]"
#puts "DEC in D M S: [deg2dms $sundec]"
#puts "[deg2dms [dms2deg 24 34 56]]"
#puts [deg2dms [get_st0 [get_jd "1978 11 13 04 34 00"]]]
puts "long=[set longitude [dms2deg -0 17 25.94]]"
set latitude [dms2deg +50 47 55]
puts "RA(Saturn)=[set ra [dms2deg 10 57 35.681]] hours"
puts "Dec(Saturn)=[set dec [expr 8 + 25/60. + 58.10/3600.]] deg"
set st [expr [get_st0 [get_jd "1978 11 13 4 34 00"]] + [dms2deg 4 34 00]*1.002737908]
puts "St0=$st hours"
set h [expr $st - $longitude - $ra]
puts "HA(Saturn)=$h"
puts "Az h = [hadec2azalt $h $dec]"
exit
