#!/bin/sh
# the next line restarts using tcl/tk interpreter\
exec wish "$0" "$@"

#
#  SUPERVISOR: Observation process manager
#
namespace eval sv {
# version:
variable version 0.11

# The only hardcoded setting is a name of configuration file:
variable cfg_file_name [pwd]/sv.cfg

# List of real components (not aliases) which need INIT,PARK etc:
variable comp_list [list]

# SV parameteres obligatory for SV functionality:
variable SV_check_list [list tmout obs_comp_list oscen cscen revive_time\
	emergency_sys]
# Component parameters obligatory for functionality:
variable comp_check_list [list host port ident]

# 
# Reading the SV configuration file (SV and components parameters)
#
# @param file Configuration file name (see its syntax in sv.cfg preamble)
# @return	number of parameters set
#
proc read_cfg file {

	if {[catch {set in [open $file r]}]} {
		#ENOCFG:
		problem "No cfg $file found" ENOCFG
	}
	# Until first "component xxx" line the parameters are those of SV:
	set comp SV
	# Declare SV as a global (array) variable:
	global SV
	set linenum 0
	set parnum 0
	while {[gets $in line] >= 0} {
		incr linenum
		# Locate possible comment character:
		if {[set ci [string first \# $line]] >=0} { 
			# remove righthand comment after ci:
			set line [string range $line 0 [expr $ci-1]]
		}
		# delete possible leading/trailing spaces:
		set line [string trim $line]	
		# locate parameter name end:
		set dp [string wordend $line 0]
		set sl [string length $line]
		if {$dp<$sl} {
			set pname [string tolower [string range $line 0 [expr $dp-1]]]
			set pvalue [string trim [string range $line \
				[string wordstart $line	$dp] end]]
			if {$pname=="component"} {
				# declaration of the next component section
				set comp $pvalue
				# new component or addition to previously defined?:
				if {[lsearch $sv::comp_list $comp]<0} {
					# add new component to the list of known components
					lappend sv::comp_list $pvalue
				}
				# declare the global array of component parameters:
				global $pvalue
				# create the same name procedure for accessing array
				set s "{ return \$::$pvalue\(\$param\) }"
				uplevel #0 proc $pvalue param $s
			} elseif {$pname=="alias"} {
				# declaration that the current component is an alias of another
				# delete this component if it is different from what it should
				# alias:
				if {[info exists ::$comp] && \
					[array get $comp *]!=[array get	$pvalue *]} {
					# Copy all elements into target of aliasing:
					array set $pvalue [array get $comp *]
					add_log "CFG: ATTENTION! $comp is alias of $pvalue but has own parameters"
					add_log "CFG: Move $comp el-s [array get $comp *] to $pvalue!"
					unset ::$comp
				}
				# delete this component from the list as well, since no 
				# personal init/stop/park is needed for it:
				set i [lsearch $sv::comp_list $comp]
				set sv::comp_list [lreplace $sv::comp_list $i $i]
				upvar #0 ::$pvalue ::$comp
			} else {
				upvar #0 $comp cmp
				set cmp($pname) $pvalue
				incr parnum
			}
		} else {
			#EBADCFG:
			if {$dp>0} {
				problem "Single word in cfg $file, not a comment (line $linenum)"\
				 EBADCFG
			}
		}
	}
	close $in
	
	# Check CFG:
	foreach p $sv::SV_check_list {
		if {![info exists SV($p)]} {
			problem "CFG file: no SV($p) found" ENOPCFG
		}
	}
	foreach c $sv::comp_list {
		foreach p $sv::comp_check_list {
			upvar #0 $c cmp
			if {![info exists cmp($p)]} {
				problem "CFG file: no $c\($p\) found" ENOPCFG
			}
		}
		# Create status and chanid elements (could be accessed before creation)
		array set $c {chanid "closed" status "UNKNOWN"}
	}
	return $parnum
}

#-------------------------- COMMAND SERVICE --------------------------------
# Array of lists of "after" results and commands, for which these
# after-timers were set, indexed by cmd_id,comp pairs: e.g. 
# cmd_array(0,OBJM) = {after#1 {GET ID}}
# After receiving a reply, the element must be unset.
variable cmd_array

# Common sv-namespace variable - command identifier
variable cmd_id 0

# For displaying the commands history tail:
variable answered_cmd_list [list]
variable answered_cmd_list_n 4

#
# Sending a command to a component.
#
# @param comp component name (as in CFG)
# @param kwd  command keyword (GET, RUN, etc)
# @param args >=0 switches or parameter=value pairs
# [@param] "&" as last argument: return after 1st reply even if it's OK WAIT=...
# @return command identifier on success, -1 on error
#

proc cmd {comp kwd args} {
	global $comp SV
	upvar #0 $comp cmp
	set kwd [string toupper $kwd]
	
	# Send a command string to the component and register and schedule its 
	# timeout:
	set chanid $cmp(chanid)
	# - chanid here is assigned in connect{}

	# Detect request for background execution command: no waiting for completion, 
	# only for the first reply:
	
	if {[set i [lsearch $args "\&"]]>=0} {
		set args [lreplace $args $i $i]
		set bkgrexec 1
	} else {
		set bkgrexec 0
	}

	set cmd "$kwd $args"
	if {$chanid=="closed" || 
		[catch {puts $chanid "$sv::cmd_id $cmd"; flush $chanid}]} {
		#ECMDDSC: ignore
		problem "Attempt to send a command to disconnected $comp" ECMDDSC \
			$comp $cmd
		return -1
	} else {
		#ECMDLOS: Register the sent command in an array
		set s [subst {{Timout waiting for $comp reply} ECMDLOS $comp {$cmd}}]
		set sv::cmd_array($sv::cmd_id,$comp) [list \
			[after $::SV(tmout) sv::problem $s] $cmd]
	}
	add_log "->$comp $sv::cmd_id $cmd"
	set prev_cmd_id $sv::cmd_id
	# Modify the command identifier according to the 0..65535,0.. cycling rule:
	if {$sv::cmd_id==65535} { set sv::cmd_id 0 } else { incr sv::cmd_id }

	# Delay the return from the command until it's replied:
	# Wait until variable is unset (if no & supplied). WAIT-reply sets it back!
	while {[info exists sv::cmd_array($prev_cmd_id,$comp)]} {
			vwait sv::cmd_array($prev_cmd_id,$comp) 
			if {$bkgrexec} break
	}
	return $prev_cmd_id
}

#
# Reply whether the command with a given identifier is under execution
#
# Searchs the array of commands for occurrence of the given command
# identifier.
# @param comid the command identifier (as returned by cmd{})
# @return 1 - the command is still under execution, 0 - command is over 
# (or was not not issued at all)
#
proc is_cmd comid {
	return [expr [llength [array names sv::cmd_array $comid,*]]>0]
}

#
# Wait until one of the command(s) execution is finished
#
# Each time the command list is updated (by cmd{}, get_reply{}), checks whether 
# one of the commands (which identifiers are provided) is over (removed
# by get_reply{} from the command list) and if so - return the identifier of the
# finished command. If several commands are over at the time of checking, the
# identifier which stands earlier in the list of arguments is returned.

# @param args one or several command identifiers
# @return command identifier (one of those in args) which command is over
#
proc wait_cmd args {
	while 1 {
		foreach c $args {
			if {![is_cmd $c]} {return $c}
		}
		vwait sv::cmd_array
	}
}

# Here we include the utilities for parsing the command/reply strings and
# composing the reply/commands from list of parameter[=value]s:
source parsecmd.tcl

#
# Common handler for all component channels reading which parses their replies
#
# This handler is called from the individual component channel handlers - the
# procedures get_reply_<componentname> created in connect{} and bound to the
# channel by the fileevent command.
# @param comp component name
# @return not defined
#
proc get_reply comp {
	global $comp
	upvar #0 $comp cmp
	if {[gets $cmp(chanid) reply] < 0} {
		#ECMPDSC: end of file condition: disactivate current handler
#		close $cmp(chanid)
		problem "Lost connection to $comp" ECMPDSC $comp
		return
	}
	add_log "<-$comp $reply"
	
	# Create reply-list from the reply-string:
	set rl [parse_command $reply]

	# Extract command identifier and keyword:
	set ci [ lindex $rl 0 ]
	set kwd [lindex $rl 1]
	# Remove COMID and KWD from reply leaving the parameter=value's only:
	set rl [lreplace $rl 0 1]
#puts "receive1: cmd_array($comp)=[array get sv::cmd_array *,$comp]"
#puts "After=[lindex $sv::cmd_array($ci,$comp) 0]"
	# Cancel the timeout timer for this command and delete it from list:
	set cmd [lindex $sv::cmd_array($ci,$comp) 1]
	if {[catch {after cancel [lindex $sv::cmd_array($ci,$comp) 0]}]} { 
		#ECMDID: if caugth then no timer for such COMID was registered for COMP
		problem "Unexpected reply (no command with ci=$ci sent to $comp)"\
		 ECMDID $comp
	}
	
	# Parse rest of reply and assign the replied quantities to the component 
	# parameters:
	
	if {[lsearch $rl switch]>=0} {
		#ECMDPAR:
		problem "Parsing reply from $comp: param. without value in $reply"\
		 ECMDPAR $comp
	} else {
		# remove wait-parameter for the case if it was a delayed result:
		array unset $comp wait
		# set the replied parameter(s) with value(s) for the component:
		array set $comp $rl
		if {[llength [array names $comp wait]]>0} {
			# Postpone final reply getting a bit more:
			set wms [expr $cmp(wait) * 1000 + $::SV(tmout)]
#puts "Postpone reply timeout by another $wms ms"
			#ECMDLOW: Register another timer of the command in an array
			set s [subst {{Timout waiting for $comp reply (after wait)}\
			 ECMDLOW $comp {$cmd}}]
			set sv::cmd_array($ci,$comp) [list 	[after $wms sv::problem $s]\
				[lindex $sv::cmd_array($ci,$comp) 1] ]
#puts "postpone: cmd_array($comp)=[array get sv::cmd_array *,$comp]"
		}
	}
	# Remove the command from the array if there was no wms set (WAIT parsed).
	# We could not do it before parsing! since there must be no period of
	# respective element absence (vwait in cmd{}).
	if {[info exists wms]==0} {
		# GUI: append the current answered command to the list and remove oldest
		lappend sv::answered_cmd_list "$ci ->$comp $cmd  <- $reply"
		if {[llength $sv::answered_cmd_list]>$sv::answered_cmd_list_n} {
			set sv::answered_cmd_list [lreplace $sv::answered_cmd_list 0 0]
		}
		array unset sv::cmd_array $ci,$comp
	}
	
	switch $kwd {
		"ERROR"
			{ 	
				problem "Command cannot be executed: $cmp(status)"\
				 ECMPSTA $comp $cmd
			}
		"OK"
			{ }
		default
			{
				problem "Unknown keyword: $reply from $comp" ECMDKWD $comp
			}
	}

#puts "receive: cmd_array($comp)=[array get sv::cmd_array *,$comp]" 
#puts "$comp params:[array get ::$comp]"
}


#------------------------- TIME SERVICE ------------------------------------
#
# Return the local time stamp according to the adopted format
#
# Time stamp is adopted in format YYYY-MM-DD HH-MM-SS and composed from the
# machine UT time to be consistent with other program log files.
# @param sec1969	Time is seconds after 1969 (Unix integer time). Current by
# 	default
# @return Time stamp string
#
proc time_stamp {{sec1969 -1}} {
	if {$sec1969==-1} {
		set sec1969 [clock seconds]
	}
	return [clock format $sec1969 -format "%Y-%m-%d %H:%M:%S" -gmt true ]
}

#------------------------ ESTABLISH COMMUNICATION ---------------------------

#
# Connect component(s) listed in arguments
#
# Tries to create the socket-connection to components using their host and port
# parameters listed in CFG. Channel is configured with default parameters. 
# If connected, creates the channel reading handler and binds the channel to it. 
# The reading handler simply calls get_reply{}.
# Then checking of the component identifier is made: error is caused if no match
# between reply and cfg name is found. Finally, component status is asked.
# @param args component identifier(s) or "all" for all existing components
#
proc connect args {
	if {$args=="all"} {set args $sv::comp_list}
	foreach comp $args {
		global $comp 
		upvar #0 $comp cmp
		add_log "Connect $comp at $cmp(host) port $cmp(port)"
		# Check if comp is started:

		if {[catch {set t [socket $cmp(host) $cmp(port)]} ]} {
			#ENOCMP:
			problem "Cannot connect to $comp" ENOCMP $comp
        } else {
			set cmp(chanid) $t
			# Create the procedure for handling the replies (Overwrites existing):
			uplevel #0 "proc get_reply_$comp {} { sv::get_reply $comp }"
			# Set it as a component channel reading handler:
			fileevent $cmp(chanid) readable get_reply_$comp
        }
		
		set ident_cfg $cmp(ident)
		set c1 [cmd $comp GET IDENT]
		if {![string match -nocase $cmp(ident) $ident_cfg]} {
			problem "$comp ident $cmp(ident) does not match $ident_cfg from cfg"\
			 ENMCMP $comp
		}
		cmd $comp GET STATUS
		# Park if not parked?
	}
}

#
# Disconnect component(s) closing its channel
#
# Channel identifier is assigned "closed" for recognition by cmd{}.
# @param args component identifier(s) or "all" for all existing comps.
# @return 1 if channel was open, 0 if channel was already closed
#
proc disconnect args {
	set res 0
	if {$args=="all"} {set args $sv::comp_list}
	foreach comp $args { 
		global $comp
		upvar #0 $comp cmp
		# the channel reading handler will be unset automatially now:
		if {$cmp(chanid)!="closed"} {
			set res 1
			add_log "Disconnect $comp"
			# Channel may appear not found if already closes: catch this error
			catch {close $cmp(chanid)}
			set cmp(chanid) closed
			set cmp(status) UNKNOWN
			# delete all pending commands to the component (to cause 
			# possible wait_cmd{} to return):
			array unset sv::cmd_array *,$comp
		}
	}
	return $res
}

#------------------------ PARKING/INITIALIZATION ---------------------------
#
# Initialize component(s)
#
# Issue INIT command to all components listed in complist and wait until all
# components finish initialization.
# @param complist name of the list with component names or "all" for all comps.
#
proc initialize complist {
	if {$complist=="all"} {
		set clist $sv::comp_list
	} else {
		upvar $complist clist
	}
	# Issue the Initialize command to all components:
	foreach comp $clist {
		lappend initcmdlist [cmd $comp INIT &]
	}
	# wait for all init commands are over:
	add_log "Wait for components to finish initialization..."
	foreach c $initcmdlist { wait_cmd $c }
	add_log "Initialization is over."	
}

#
# Stop and park components
#
# Issue STOP NOW and PARK commands and wait until all component finish parking.
# @param complist name of the list with component names or "all" for all comps.
#
proc stop_park complist {
	if {$complist=="all"} {
		set clist $sv::comp_list
	} else { 
		upvar $complist clist
	}
	add_log "Stop and park components $clist"
	set parkcmdlist [list]
	foreach comp $clist {
		upvar #0 $comp cmp
		if {$cmp(chanid)=="closed"} continue
		cmd $comp STOP NOW
		if {$cmp(status)=="BUSY"} {
			#ECMPNST: 
			problem "$comp is busy after STOP NOW" ECMPNST $comp
		}
		if {[is_cmd [set p [cmd $comp PARK &]]]} {lappend parkcmdlist $p}
	}
	# wait for all park commands are over:
	if {[llength $parkcmdlist]} {
		add_log "Wait for components to finish parking..."
		foreach c $parkcmdlist { wait_cmd $c }
		add_log "Parking finished."
	}

}

#------------------------ SCENARIA MANAGEMENT -------------------------------
#
# Check the scenaria (for observations, monitoring)
#
# The checking of scenaria existence is performed. Currently that's all.
# @param args list of scenaria file names
# @return 1 if all scenaria are Ok, 0 otherwise
#
proc check_scen args {
	foreach s $args { 
		if {![file readable $s]} {
			problem "Scenario $s is not readable" EBADSCE
			return
		}
	}
	add_log "Scenaria $args are Ok"
}

# Flag of observations going on. Precisely, this means that safe Tcl interpreter
# is created and executes the observations scenario (see start_obs{}).
variable is_observations_now 0

#
# Return flag of "observations now" for calling scenario.
#
# @return $is_observations_now
#
proc is_observations_now {} {return $sv::is_observations_now}

#
# Update idle callbacks for refreshing the SV window
#
# This utility makes updates of idle tasks (mainly GUI) with a delay
# periodicity. Once started, it works via delayed recursion while observations 
# are under way. So it should be started in start_obs{}.
# The reason to do so is that scenario may cause interface frozen.
#
# @param delay update periodicity in milliseconds.
#
proc update_idle delay { 
	update idletasks
	if {$sv::is_observations_now} {	after $delay sv::update_idle $delay }
}

# Tcl interpreter identifier for running observation scenario
variable tclobs 0
# flag of cancellation of observations BEFORE SV returns:
variable disable_start_obs false

#
# Start observations scenario
#
# Create a safe interpreter (for network safety reasons etc.) and call 
# initialize{} of all components. Then, if initialize{} was not interrupted by 
# stop_obs{} (i.e. interpreter still exists), create the aliases 
# of needed procedures for obs. scenario and source scenario with the created 
# interpreter on the background (after 10ms delay). En fin, start update_idle{}.
# NOTE that interaction of start_obs{} and stop_obs{} is a FINE PLACE in SV
# programming!
#
proc start_obs {} {
	global SV errorCode
	# detect whether the stop_obs was forced before with disable_start_obs==true
	if {$sv::disable_start_obs} return
	# avoid starting observations during initialization (i.e. reentered)"
	set sv::disable_start_obs true
	# determine which components are not initialized (by their status)
	# (defined at least in connect{}). Circ.monitor MUST initialize components
	# which it needs! - namely METEO and CIRC.
	set sv::tclobs [interp create -safe]
	sv::initialize SV(obs_comp_list)

	# make aliases for accessing commands in the interpreter running monitor. 
	# first check whether observations are still started (sv::initialize{} may
	# be interrrupted!):
	if {[interp exists $sv::tclobs]} {
		sv::add_log "Start observations: interpreter $sv::tclobs"
		foreach c {cmd is_cmd wait_cmd} {$sv::tclobs alias $c $c}		
		foreach c $sv::comp_list { $sv::tclobs alias $c $c}
	# below catch catches the eval of deleted interp. commands when stopped obs.
		after 10 {
			add_log "Source $SV(oscen)"
			if {[catch {$sv::tclobs invokehidden -global source $SV(oscen)}\
				  obsErr]} {
				  if {[lindex $errorCode 0]!="CORE" ||\
				   [lindex $errorCode 1]!="IDELETE"} {sv::problem $obsErr EOBSSCE}
			}
			add_log "$SV(oscen) sourcing is canceled"
		}
	}
	update_idle 2000
	set sv::is_observations_now 1
	set sv::disable_start_obs false
}

#
# Stop observations 
#
# Deletes the safe interpreter for observations scenario created in start_obs{}
# thus stopping observations (next command in obs.scen fails with an error
# "attempt to call a command in deleted interpreter" which is caught in a script
# started from start_obs{} on a background. NOTE that this is a FINE POINT in
# such a programming of SV!
# After deletion of interpreter, stop_park{} is called for components
# initialized in start_obs{}.
#
# @param disable_start_obs flag of protection of immediate restart of
#        observations, false by default. Used for manual stopping.
#
proc stop_obs {{disable_start_obs false}} {
	global SV
	if {![interp exists $sv::tclobs]} return
	set sv::disable_start_obs $disable_start_obs
	sv::add_log "Stop observations: stop interpreter $sv::tclobs"	
	interp delete $sv::tclobs
	set sv::tclobs 0
	set sv::is_observations_now	0
	sv::stop_park SV(obs_comp_list)
}

# Tcl interpreter identifier for running the monitor scenario
variable tclmon 0
#
# Start circumstance monitor scenario
#
# Similarly to start_obs{}, creates the safe interpreter, makes aliases of
# (wider scope of) commands and starts the monitor scenario in this interpreter
# on a background.
#
proc start_monitor {} {
	global SV
	add_log "Start monitor $SV(cscen)..."
	set sv::tclmon [interp create -safe]
	# make aliases for accessing commands in the interpreter running monitor:
	foreach c {cmd is_cmd wait_cmd start_obs stop_obs add_log is_observations_now} {
		$sv::tclmon alias $c $c
	}
	foreach c $sv::comp_list { $sv::tclmon alias $c $c}
	after 10 {
		add_log "Source $SV(cscen)"
		if {[catch {$sv::tclmon invokehidden -global source $SV(cscen)}\
			  cirErr]} {
			  if {[lindex $errorCode 0]!="CORE" ||\
			   [lindex $errorCode 1]!="IDELETE"} {sv::problem $cirErr ECIRSCE}
		}
		add_log "$SV(cscen) sourcing is stopped"
	}
	set sv::is_observations_now 0
}

#
# Stop circumstance monitor:
#
proc stop_monitor {} {
	global SV
	if {![interp exists $sv::tclmon]} return
	sv::add_log "Stop monitor: stop interpreter $sv::tclmon"	
	interp delete $sv::tclmon
	set sv::tclmon 0
}
#-------------------------- INTERFACE, LOGGING ------------------------------
# Log file descriptor
variable logfileid 0
#
# Open a log-file of the session as YYMMDDsv.log, where date is UTC date
#
proc open_log {} {
	set sv::logfileid [open [clock format [clock seconds] \
		-format "%y%m%dsv.log" -gmt true ] a]
	add_log "Supervisor V$sv::version session open"
}
#
# Close log-file
#
proc close_log {} {
	if {$sv::logfileid!=0} {close $sv::logfileid; set sv::logfileid 0 }
}
#
# Add a log line in the sv.log and on a console with a time stamp.
#
proc add_log text {
	puts "[time_stamp] $text"
	if {$sv::logfileid!=0} {
		puts $sv::logfileid "[time_stamp] $text"
		flush $sv::logfileid
	}
}
#
# Handle errors 
#
# Procedure takes some actions for correction the problem in situ and returns
# the control back to calling procedure - if it is possible; or issues a Tcl
# error to avoid return of the control and cancel the normal SV work with help
# of catch construction in main{}. Description of the error handling policy if
# in error.lst.
#
# @param message Problem error message
# @param code error code for automatic deciphering
# Optional arguments which are supplied and used depending on the error
# and action which may be performed to handle situation:
# @param comp	Component name
# @param cmd	Command sent
#
proc problem {message code {comp NO} {cmd NO}} {
	global SV argv errorInfo
	add_log "ERROR: $message"
	switch -glob $code {
		ECMDDSC return

		E*CFG - EBADSCE {	
			# Startup errors: simply stop SV
			exit
		}
		
		EN*CMP {
			terminate
			exit
		}
		
		EOBSSCE - ECIRSCE {
			# Scenaria errors: stop with messages:
			add_log $errorInfo
			exec $SV(emergency_sys) &
			terminate
		}
		
		ECMDID - ECMPNST - ECMDPAR - ECMDKWD -		
		ECMDLO* - ECMPDSC - ECMPFAT {
			# if connection is lost - first close channel:
			if {[string match ECMDLO* $code] || \
				[string match ECMPDSC $code]} {
				# if channel is already closed (secondary error) - return
				if {![disconnect $comp]} return
			} else {
				# otherwise - park bad component before doing other things:
				stop_park $comp
				disconnect $comp
			}
			
			exec $SV(emergency_sys) &
			
			if {[info exists ::${comp}(optional)]} {
				add_log "Component $comp is optional, so we closed it"
			} else {
				# Component is essential for normal work: try to revive:
				terminate
				add_log "Shell will sleep for $SV(revive_time) sec and then SV restarts"
				exec echo "/bin/sleep $SV(revive_time)\n./sv.tcl $argv &\n\
				rm sv_again" >sv_again
				exec chmod u+x sv_again
				exec ./sv_again &
				exit
			}
		}
		
		ECMPSTA {
			upvar #0 $comp cmp
			if {$cmp(status)=="BUSY"} {
				add_log "wait while $comp is $cmp(status)..."
				vwait ::${comp}(status)
				add_log "Ok: $comp is $cmp(status): send $cmd to $comp again"
				cmd $comp $cmd
			} else {
				add_log "IGNORED"
			}
		}
		
		default {
			# Here we are if we missed some (new?) error code:
			exec $SV(emergency_sys) &
			error $message "Unexpected problem code"
		}
	}
	return
}

#
# Create the GUI for monitoring the SV and components state. Supports
# configuration from SV cfg file ("show"-lists of variables for monitoring);
# component names are the buttons which allow to issue the command to the given
# component.
#
proc create_window {} {
	package require Tix
	wm withdraw .
	set w [toplevel .t]
	wm title .t "Supervisor v$sv::version"
	wm protocol .t WM_DELETE_WINDOW sv::terminate
	
	# create frame-table with sub-frames-columns each for its component items
	set f [frame $w.ft]
	pack [label $f.tlabel -text Components] -side top
	pack $f -fill x -padx 2 -pady 2
	
	set f [frame $w.fcmp -borderwidth 3]
	foreach c $sv::comp_list { 
		global $c
		upvar #0 $c cmp
		set fc [frame $f.f$c -bd 3 -width 100 -relief groove]
		# button with component name:
		menubutton $fc.cn -text $c -state normal -menu $fc.cn.menu
		set m [menu $fc.cn.menu]
#		set cm  "set ${c}(status) ISOLATED"
#		$m add command -label Isolate -command $cm
#		set cm "cmd $c get status"
#		$m add command -label Activate -command $cm
		$m add command -label Command -command "sv::cmd_dialog $c"
		label $fc.status -textvariable ${c}(status) -fg red
		pack $fc.cn $fc.status -fill x -padx 2 -pady 2 -side top
		if {[info exists cmp(show)]} {
			foreach v $cmp(show) {
			if {$v=="data"||$v=="status"} continue
				frame $fc.v$v
				pack [label $fc.v$v.lab -text "$v:" ] -side left
				pack [label $fc.v$v.val -textvariable ${c}($v) -fg blue] -side \
					left
				pack $fc.v$v -anchor w
			}
		}
		pack $fc -side left -anchor n -expand true -fill y
	}
	pack $f -pady 2 -expand true
	
	# Create a section with components data lines:
	set f [frame $w.fdata]
	foreach c $sv::comp_list {
		global $c
		upvar #0 $c cmp
		if {[info exists cmp(show)]==0 || [lsearch $cmp(show) data]<0} continue
		set fc [frame $f.f$c]
		pack [label $fc.dlabel$c -text "$c DATA:"] -side left -anchor w
		pack [label $fc.data$c -textvariable ${c}(data) -fg blue] -side left -anchor w
		pack $fc -padx 2 -pady 2 -anchor w
	}
	pack $f -pady 2 -anchor w

	# Create the message with observations/monitoring state:
	set f [frame $w.obs]
	pack [label $f.lab -text "Activity:" ] -side left
	pack [label $f.val -text "undefined" -fg blue] -side left
	pack $f -pady 2
	
	
	# Create the window with the last and delayed-execution commands:
	set f [frame $w.cmd -takefocus 0]
	pack [label $f.label -text "Last commands:"]
	pack [listbox $f.cmdlist -takefocus 0 -selectmode none -height \
		[expr [llength $sv::comp_list]+$sv::answered_cmd_list_n]  -width 60]
	pack $f -padx 2 -pady 2 -anchor w
	trace variable sv::cmd_array wu sv::update_cmd_array
	
	# Commands to manipulate observations state:
	set f [frame $w.obscmd]
	pack [button $f.stop -text "Stop Observations" -state disabled \
		-command {sv::stop_obs true}] -side left -padx 10
	pack [button $f.enable -text "Allow Observations" \
		-command {set sv::disable_start_obs false}]  -side left -padx 10
	pack [button $f.terminate -text "Terminate" \
		-command sv::terminate]  -side left -padx 10
	pack $f -pady 2
	trace variable sv::is_observations_now w sv::update_obs_buttons
	trace variable sv::tclobs w sv::update_obs_buttons

	# Commands to manipulate circumstance monitor:
	set f [frame $w.moncmd]
	pack [button $f.stop -text "Stop monitor" -state disabled \
		-command {sv::stop_monitor}] -side left -padx 10
	pack [button $f.start -text "Start monitor" \
		-command {sv::start_monitor}]  -side left -padx 10
	pack $f -pady 2
	trace variable sv::tclmon w sv::update_mon_buttons

}

#
# Depending on $is_observations_now, activate or disable buttons in GUI
#
# This procedure serves to reflect the modifications of is_observations_now
# variable in the SV window (see "man n trace"). Also, the "preparation" stage
# is detected when tclobs interpreter is already created while is_observations_now
# is still false.
# @param name1 accessed variable name
# @param name2 element name if name1 is an array
# @param op access operation code (w - write, r - read, u - unset; w normally)
#
proc update_obs_buttons {name1 name2 op} {
	if {$sv::is_observations_now} {
		.t.obscmd.stop configure -state active
		.t.obs.val configure -text "Observations"
	} elseif {$sv::tclobs!=0} {
		.t.obs.val configure -text "Preparing observations..."
	} elseif {$sv::tclmon==0} {
		.t.obscmd.stop configure -state disabled	
		.t.obs.val configure -text "Doing nothing..."
	} else {
		.t.obscmd.stop configure -state disabled	
		.t.obs.val configure -text "Monitoring conditions"
	}
}
proc update_mon_buttons {name1 name2 op} {
	if {$sv::tclmon!=0} {
		.t.moncmd.stop configure -state active	
		.t.moncmd.start configure -state disabled	
	} else {
		.t.moncmd.stop configure -state disabled	
		.t.moncmd.start configure -state active
	}
}

variable tix_combo_box {}
variable label_cont ""
#
# Operator's dialog with a component from Tk window:
# @param comp Component name
#
proc cmd_dialog comp {
	set t [toplevel .dia$comp]
	wm title $t "SV command dialog with $comp"
	set sv::label_cont "Enter command to $comp"
	set f [tixComboBox $t.cmd -selectmode browse -editable true -dropdown true\
		-variable sv::tix_combo_box -padx 20 -pady 4\
		-labelside top -fancy true -history true -historylimit 20 \
		-command "sv::cmd_issue $comp" ]
	$t.cmd subwidget label configure -width 100 -textvariable sv::label_cont
	foreach cmnd { INIT RUN STOP PARK FREE GET SET } { $t.cmd insert end $cmnd } 
	pack $t.cmd
}

#
# Send a command to the component and put a reply in the sv::label_cont var.
# @param comp Component name
# @param cmd command keyword and arguments (if needed)
#
proc cmd_issue {comp cmd} {
	cmd $comp $cmd
	set reply [lindex $sv::answered_cmd_list end]
	set sv::label_cont "Reply is: $reply\n New command:"
}

#
# Update the view of the commands list in the SV window: invoked by trace var.
# First creates the new list of delayed commands, then deletes the old
# displayed list and adds the sorted concatenation of the delayed commands and
# answered commands
#
proc update_cmd_array {name1 name2 op} {
	set cl [list]
	foreach n [array names sv::cmd_array] {
		set s "$n [lindex $sv::cmd_array($n) 1]"
		set s [string replace $s [set i [string first , $s]] $i " ->"]
		lappend cl $s
	}
	set cls [lsort -dictionary [concat $cl $sv::answered_cmd_list]]
	.t.cmd.cmdlist delete 0 [expr [.t.cmd.cmdlist size] - 1]
	foreach e $cls {
		.t.cmd.cmdlist insert end $e}
}

#
# Terminate SV
#
# Stops monitor interpreter (if it still executes something directly), stop
# observations (if under way) and do all the job to stop SV correctly.
# Procedure is not reenterable, so the protection is set against reentering.
#
variable terminating false
proc terminate {} {
	add_log "TERMINATE"
	stop_monitor
	if {[interp exists $sv::tclobs]} {stop_obs true}
	trace vdelete sv::cmd_array wu sv::update_cmd_array
	trace vdelete sv::is_observations_now w sv::update_obs_buttons
	trace vdelete sv::tclobs w sv::update_obs_buttons
	trace vdelete sv::tclmon w sv::update_mon_buttons
	stop_park all
	disconnect all
	close_log
	destroy .t
}

###################### Starting supervisor:
#
# Starting the normal algorithm of SV functionality
#
proc main {} {
	open_log
	read_cfg $sv::cfg_file_name
	create_window
	connect all
	global SV
	check_scen $SV(cscen) $SV(oscen)
	# arrival of TERM signal will remove the SV window causing tkwait...t to end
	# this demands TclX package to be loaded:
	if {[catch {package require Tclx}]} {
		tk_messageBox -type ok -parent .t -icon warning -message \
		"To enable remote correct SV kill-termination, install TclX package"
	} else {
		signal trap SIGTERM { sv::terminate } 
	}
	start_monitor
	tkwait window .t
}

namespace export -clear cmd is_cmd wait_cmd start_obs stop_obs add_log is_observations_now

}

# import those commands of namespace sv which are used in scenaria:
namespace import sv::cmd sv::is_cmd sv::wait_cmd sv::start_obs sv::stop_obs sv::add_log sv::is_observations_now

sv::main
exit
